/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: UNF_SCHK.C $
   $Locker: ska $	$Name:  $	$State: Exp $

	Minimum checking of given NLS data

	Checks for:
	+ Missing standard tables 1, 2, 4, 5, 6, 7,& 0x23
	+ table size mismatches (of standard tables only)
	+ '\0' terminated strings in table 1 and dateFmt etc.
	+ DBCS 0 terminator word

	Return:
		The number of flaws

*/

#if defined(__TURBOC__) || defined(__WATCOMC__)
#include <stddef.h>
#endif

#include "unf.loc"

#ifdef _MICROC_
#endif
#include <string.h>
#include <stdlib.h>

#include "unf_post.loc"

#ifdef RCS_Version
static char const rcsid[] = 
	"$Id: UNF_SCHK.C 1.2 2000/06/17 22:32:12 ska Exp ska $";
#endif

#define unf_error ++flaws; (*error_fct)( NULL, 

#define numOK(a,expr)											\
	if(cntry->a expr) {											\
		unf_error UNF_EValue, "%s value of table %u invalid", #a , 1);	\
	}

#define hasTerm(a)											\
	if(memchr(cntry->a, 0, sizeof(cntry->a)) == NULL) {		\
		unf_error UNF_EStringValue, "%s value of table %u not '\\0' terminated", #a , 1);		\
	}

int unf_sanityCheck(struct UNF_DATA *d, UNF_Error_Fct *efct)
{	int flaws;					/* number of problems */
	UNF_Error_Fct *error_fct;	/* function to report errors to */
	int *tab;		/* offset of tables */
	int i, fct;

	if((tab = malloc(256 * sizeof(int))) == NULL) {
		unf_error UNF_EOutOfMemory, "Out of memory");
		return 1;
	}
		/* fill all (int) values with -1 */
	memset(tab, -1, 256 * sizeof(int));
	flaws = 0;
	error_fct = efct? efct: (UNF_Error_Fct*)unf_dummy_error;

	if(!d) {
		unf_error UNF_ESyntax, "No NLS package supplied");
		goto errRet;
	}

	/* Check country ID */
	switch(d->country) {
	case (UWORD)UNF_DEFAULT:
		unf_error UNF_EMissingBasedata, "Missing country ID or codepage");
		break;
#if UNF_DEFAULT != 0
	case 0:
#endif
#if UNF_DEFAULT != -1 && UNF_DEFAULT != 0xffff
	case 0xffff:
#endif
		unf_error UNF_EInvalidBasedata, "Invalid country ID or codepage");
		break;
	}
	/* Check codepage ID */
	switch(d->codepage) {
	case (UWORD)UNF_DEFAULT:
		unf_error UNF_EMissingBasedata, "Missing country ID or codepage");
		break;
#if UNF_DEFAULT != 0
	case 0:
#endif
#if UNF_DEFAULT != -1 && UNF_DEFAULT != 0xffff
	case 0xffff:
#endif
		unf_error UNF_EInvalidBasedata, "Invalid country ID or codepage");
		break;
	}

	/* Check existance of tables */
	if(!d->numData) {
		unf_error UNF_ENoTables, "No single table specified");
		goto errRet;
	}
	i = d->numData;
	while(i--) if(d->data[i]) {
		if(tab[fct = d->data[i]->fct] != -1) {
			unf_error UNF_EDuplicateTable, "Duplicated table %u", fct);
		} else {
			tab[fct] = i;
		}
	}

	/* Checking table 1 Ext Cntry Info */
	if(tab[1] == -1) {
		unf_error UNF_EMissingTable, "Missing required table %u", 1);
	} else if(d->data[tab[1]]->size < sizeof(struct nlsExtCntryInfo) - 1) {
		unf_error UNF_ESizeTable, "Size of table %u is invalid", 1);
	} else {
		struct nlsExtCntryInfo *cntry;

		cntry = (struct nlsExtCntryInfo *)(&d->data[tab[1]]->data[-1]);

		/*numOK(subfct, != 1) UNF does NOT contain the subfct number */
		numOK(size, > d->data[tab[1]]->size - 2)
		numOK(countryCode, != d->country)
		numOK(codePage, != d->codepage)
		numOK(dateFmt, > 2)
		hasTerm(curr)
		hasTerm(thSep)
		hasTerm(point)
		hasTerm(dateSep)
		hasTerm(timeSep)
		numOK(timeFmt, > 1)
		hasTerm(dataSep)
	}

	/* Check table 2 upcase */
	if(tab[2] == -1) {
		unf_error UNF_EMissingTable, "Missing required table %u", 2);
	}
	/* Check table 4 filename upcase */
	if(tab[4] == -1) {
		unf_error UNF_EMissingTable, "Missing required table %u", 4);
	}
	/* Check table 5 filename terminator */
	if(tab[5] == -1) {
		unf_error UNF_EMissingTable, "Missing required table %u", 5);
	} else if(d->data[tab[5]]->size < offsetof(struct nlsFnamTerm, separators)) {
		unf_error UNF_ESizeTable, "Size of table %u is invalid", 5);
	} else {
		struct nlsFnamTerm *fn;

		fn = (struct nlsFnamTerm *)(d->data[tab[5]]->data);
		if(fn->size > d->data[tab[5]]->size - 2
		 || fn->size != offsetof(struct nlsFnamTerm, separators) - 2 + fn->numSep) {
			unf_error UNF_EValue, "%s value of table %u invalid", "size" , 5);	\
		}
		if((unsigned char)fn->firstCh > (unsigned char)fn->lastCh) {
			unf_error UNF_ERangeValue, "%s and %s of table %u define invalid range", "firstCh", "lastCh", 5);
		}
		if((unsigned char)fn->firstExcl > (unsigned char)fn->lastExcl) {
			unf_error UNF_ERangeValue, "%s and %s of table %u define invalid range", "firstExcl", "lastExcl", 5);
		}
	}

	/* Check table 6 collating sequence */
	if(tab[6] == -1) {
		unf_error UNF_EMissingTable, "Missing required table %u", 6);
	}

	/* Check table 7 DBCS */
	if(tab[7] == -1) {
		unf_error UNF_EMissingTable, "Missing required table %u", 7);
	} else if(d->data[tab[7]]->size < 4) {
		unf_error UNF_ESizeTable, "Size of table %u is invalid", 7);
	} else {
		UWORD *dbcs;

		dbcs = (UWORD*)(d->data[tab[7]]->data);
#if defined(DBCS)
		if (*dbcs != 0) {	/* DBCS environment */
			if (*dbcs < 2 || (*dbcs & 1) || *dbcs > d->data[tab[7]]->size) {
				unf_error UNF_ESizeTable, "Size of table %u is invalid", 7);
			}
			if (*(UWORD *)((UBYTE *)dbcs + *dbcs) != 0) {
				unf_error UNF_EValue, "%s value of table %u invalid", "0 terminator word" , 7);
			}
		} else	/* SBCS environment */
		if (dbcs[1] != 0) {
			unf_error UNF_EValue, "%s value of table %u invalid", "0 terminator word" , 7);
		}
#else
		if(*dbcs > 32766 || *dbcs * 2 > d->data[tab[7]]->size - 4) {
			unf_error UNF_ESizeTable, "Size of table %u is invalid", 7);
		} else if(dbcs[*dbcs + 1] != 0) {	/* mandory 0 word */
			unf_error UNF_EValue, "%s value of table %u invalid", "0 terminator word" , 7);
		}
#endif
	}

	/* Checking table 0x23 Yes/No */
#if defined(DBCS)
	if(!tab[0x23]) {
		unf_error UNF_EMissingTable, "Missing required table %u", 0x23);
	} else if(d->data[tab[0x23]]->size != sizeof( ((struct nlsPackage *)0)->yeschar) + sizeof(((struct nlsPackage *)0)->nochar)) {
		unf_error UNF_ESizeTable, "Size of table %u is invalid", 0x23);
	}
#else
	if(!tab[0x23]) {
		unf_error UNF_EMissingTable, "Missing required table %u", 0x23);
	} else if(d->data[tab[0x23]]->size != 2) {
		unf_error UNF_ESizeTable, "Size of table %u is invalid", 0x23);
	}
#endif

errRet:
	free(tab);
	return flaws;
}
